#
# odb2gomxml v0.1 / 19.06.2019
# ============================
# by T.Moeller (t.moeller@gom.com)
# GOM mbH (http://www.gom.com)
#
# ABAQUS python script to export odb-model and -displacement data
# to a *.xml file which can be read by GOM ARAMIS Professional
# to compare ABAQUS data to optical strain measurement data.
#
# It exports all steps from all *.odb-files in the current path.
#
# Usage:
# ------
# - Command line:
#   abaqus python odb2gomxml.py
#
# - ABAQUS/CAE:
#   Make sure the path-variable in this script (l. 45) is set right.
#   File -> Run Script -> Select this script in the Run Script dialog box
#
# Requirements:
# -------------
# - The *.odb-file must contain the node displacements.
# - Refer to the Abaqus Analysis User's Manual Vol.1
#   4.1.3 Output to the output database
# - Enable it in the *.inp-file by:
#   *OUTPUT,FIELD
#   *NODE OUTPUT
#   U
#
# - or enable it in ABAQUS/CAE by:
#   Step module: field output request editor   
#

import odbAccess as odba
from abaqusConstants import DOUBLE_PRECISION
import numpy as np
import os
from glob import glob
from codecs import open
import io
import base64
import struct

PATH = '.' # path to search for *.odb-files to export

def getNIdsFrom3NElement (element):
    return [list (element.connectivity)]

def getNIdsFromFlat4NElement (element):
    c = element.connectivity
    return [[c[0], c[1], c[2]],
            [c[0], c[2], c[3]]]

def getNIdsFromFlat6NElement (element):
    raise ValueError, 'Not implemented yet'

def getNIdsFromFlat8NElement (element):
    raise ValueError, 'Not implemented yet'

get_nids_from_element = {'S3':getNIdsFrom3NElement,
                         'S4':getNIdsFromFlat4NElement,
                         'S3R':getNIdsFrom3NElement,
                         'S4R':getNIdsFromFlat4NElement,
                         'CPS3':getNIdsFrom3NElement,
                         'CPS4':getNIdsFromFlat4NElement,
                         'CPS3R':getNIdsFrom3NElement,
                         'CPS4R':getNIdsFromFlat4NElement,
                         'CPE3':getNIdsFrom3NElement,
                         'CPE4':getNIdsFromFlat4NElement,
                         'CPE3R':getNIdsFrom3NElement,
                         'CPE4R':getNIdsFromFlat4NElement,
                         'SC6R':getNIdsFromFlat6NElement, 
                         'SC8R':getNIdsFromFlat8NElement}

def cleanNotSupportedMeshEffects (tris):
    double_tris = {}
    same_edges = {}
    multi_edges = {}
    valid_tris = np.ones ((tris.shape[0], ), dtype=int)
    edge_ids = np.array ([[0, 1], [1, 2], [2, 0]], dtype=int)
    for tc, tri in enumerate (tris):
        # check for doubled triangles
        try:
            double_tris[tuple (tri)]
        except KeyError:
            double_tris[tuple (tri)] = 1
        else:
            valid_tris[tc] = 0
            continue
        # check for edges used several time in the same direction (normal flip)
        tri_edges = tri[edge_ids]
        for tri_edge in tri_edges:
            try:
                same_edges[tuple (tri_edge)]
            except KeyError:
                same_edges[tuple (tri_edge)] = 1
            else:
                valid_tris[tc] = 0
        if not valid_tris[tc]:
            continue
        # check for edges used more than 2 times
        tri_edges = np.sort (tri_edges, axis=-1)
        for tri_edge in tri_edges:
            try:
                multi_edges[tuple (tri_edge)]
            except:
                multi_edges[tuple (tri_edge)] = 1
            else:
                multi_edges[tuple (tri_edge)] += 1
                if multi_edges[tuple (tri_edge)] > 2:
                    valid_tris[tc] = 0
    return tris[valid_tris.astype (bool)]

def writeGomXmlSc (nodes, triangles, fid=None, name='Component'):
    if fid is None:
        fid = open (os.path.join (PATH, '%s.xml'%(name)), 'w', encoding='utf-8')
    elif isinstance (fid, str):
        fid = open (fid, 'w', encoding='utf-8')
    elif isinstance (fid, io.TextIOWrapper):
        pass
    else:
        raise ValueError ('Unknown type of parameter fid.')
    fid.write ('''<?xml version="1.0" ?>\n<gom>\n  <header>\n    <version>2.4</version>\n    <length_unit>mm</length_unit>\n    <angle_unit>deg</angle_unit>\n  </header>\n  <measured>\n    <surface_component name="%s">\n      <triangles chunk="0">'''%(name))
    tri_text = io.BytesIO ()
    tri_text.write (struct.pack ('I', 1)) # Version
    tri_text.write (struct.pack ('I', triangles.shape[0]))
    for tri in triangles:
        tri_text.write (struct.pack ('III', tri[0], tri[1], tri[2])) # node indexes
    fid.write (base64.b64encode (tri_text.getvalue ()).decode ('ascii'))
    tri_text.close ()
    fid.write ('''</triangles>\n      <state>ok</state>\n      <geometry>\n        <surface chunk="0">''')
    sur_text = io.BytesIO ()
    sur_text.write (struct.pack ('I', 1)) # Version
    bb_min_x = np.nanmin (nodes[:, 0]) # bounding box min x
    bb_min_y = np.nanmin (nodes[:, 1]) # bounding box min y
    bb_min_z = np.nanmin (nodes[:, 2]) # bounding box min z
    bb_max_x = np.nanmax (nodes[:, 0]) # bounding box max x
    bb_max_y = np.nanmax (nodes[:, 1]) # bounding box max y
    bb_max_z = np.nanmax (nodes[:, 2]) # bounding box max z
    if bb_max_x - bb_min_x < 1.e-9: # for meshes flat in x
        bb_max_x += 1.e-9
    if bb_max_y - bb_min_y < 1.e-9: # for meshes flat in y
        bb_max_y += 1.e-9
    if bb_max_z - bb_min_z < 1.e-9: # for meshes flat in z
        bb_max_z += 1.e-9
    sur_text.write (struct.pack ('ddd', bb_min_x, bb_min_y, bb_min_z)) # bounding box min
    sur_text.write (struct.pack ('ddd', bb_max_x, bb_max_y, bb_max_z)) # bounding box max
    sur_text.write (struct.pack ('I', nodes.shape[0])) # number of points
    for x, y, z in nodes:
        if not np.isnan (x):
            sur_text.write (struct.pack ('B', 1)) # valid
            sur_text.write (struct.pack ('III', int (((x - bb_min_x) / (bb_max_x - bb_min_x)) * (2**32 - 1)),
                                            int (((y - bb_min_y) / (bb_max_y - bb_min_y)) * (2**32 - 1)),
                                            int (((z - bb_min_z) / (bb_max_z - bb_min_z)) * (2**32 - 1)))) # coordinate with integers relative to the bounding box
        else:
            sur_text.write (struct.pack ('B', 0)) # not valid
    fid.write (base64.b64encode (sur_text.getvalue ()).decode ('ascii'))
    sur_text.close ()
    fid.write ('''</surface>\n      </geometry>\n      <result/>\n    </surface_component>\n  </measured>\n  <nominal/>\n</gom>''')
    fid.close()

for odbf in glob (os.path.join (PATH, '*.odb')):
    print '-------\nhandling output database %s'%(odbf)
    odb = odba.openOdb (odbf)
    tris = {}
    nnode_ids = {}
    node_ids = {}
    nodes = {}
    displs = {}
    insts = odb.rootAssembly.instances
    for inst_name, inst in insts.items ():
        print 'reading elements from instance %s'%(inst_name)
        tris[inst_name] = []
        node_ids[inst_name] = []
        nodes[inst_name] = []
        for element in inst.elements:
            try:
                faces_node_ids = get_nids_from_element[element.type] (element)
            except KeyError:
                continue
            tris[inst_name].extend (faces_node_ids)
        tris[inst_name] = cleanNotSupportedMeshEffects (np.array (tris[inst_name]))
        nnode_ids[inst_name] = {v:0 for v in tris[inst_name].flat} # needed nodes
        if nnode_ids[inst_name]:
            print 'reading nodes from instance %s'%(inst_name)
            for node in inst.nodes:
                try:
                    nnode_ids[inst_name][node.label]
                except KeyError:
                    continue
                node_ids[inst_name].append (node.label)
                nodes[inst_name].append (node.coordinates)
        else:
            print 'no known element type found in instance %s'%(inst_name)
            continue
        nnode_ids[inst_name] = {node_id: i for i, node_id in enumerate (node_ids[inst_name])}
        nodes[inst_name] = np.array (nodes[inst_name])
        tris[inst_name] = np.searchsorted (node_ids[inst_name], tris[inst_name])
        displs[inst_name] = np.zeros (nodes[inst_name].shape)
    for step_name, step in odb.steps.items ():
        print 'handling step %s'%(step_name)
        for fc in range (len (step.frames)):
            print 'handling frame %u'%(fc)
            frame = step.frames[fc]
            outp = frame.fieldOutputs
            if outp.has_key ('U'):
                print 'get nodal deformations'
                u = outp['U']
                for v in u.values:
                    try:
                        nnode_ids[v.instance.name][v.nodeLabel]
                    except KeyError:
                        continue
                    except AttributeError:
                        continue
                    if v.precision == DOUBLE_PRECISION:
                        dv = v.dataDouble
                    else:
                        dv = v.data
                    displs[v.instance.name][nnode_ids[inst_name][v.nodeLabel], 0] = dv[0]
                    displs[v.instance.name][nnode_ids[inst_name][v.nodeLabel], 1] = dv[1]
                    # take 2D projects into account
                    try:
                        displs[v.instance.name][nnode_ids[inst_name][v.nodeLabel], 2] = dv[2]
                    except IndexError:
                        pass
            else:
                print 'no nodal deformations found in %s, step %s, frame %u\nUse "*OUTPUT,FIELD" and "*NODE OUTPUT", "U"'%(odbf, step_name, fc)
                continue
            for inst_name, inst in insts.items ():
                print 'write *.xml-file for %s, %s, step %s, frame %u'%(odbf, inst_name, step_name, fc)
                writeGomXmlSc (nodes[inst_name] + displs[inst_name], tris[inst_name], os.path.join (PATH, '%s_%s_%s_%04u.xml'%('.'.join (os.path.basename (odbf).split ('.')[:-1]), inst_name, step_name, fc)), inst_name)
print 'finished'